"use strict";exports.id=4208,exports.ids=[4208],exports.modules={"./src/commands/update.ts"(t,e,n){n.r(e),n.d(e,{handleUpdate:()=>c});var r=n("./src/analytics.ts"),s=n("./src/commands/update-core.ts");function c(t,e){return n=this,c=arguments,a=function*(t,e,n="agent-cli"){try{const c=e.get();"static"===c.channel&&(process.stdout.write("Skipping update check for static channel\n"),process.exit(0)),yield(0,s.updateCursorAgent)({dashboardClient:t,showProgress:!0,channel:c.channel,isAutoUpdate:!1,product:n}),yield(0,r.PO)(),process.exit(0)}catch(t){throw new Error(`Update failed: ${t instanceof Error?t.message:String(t)}`)}},new((o=void 0)||(o=Promise))((function(t,e){function r(t){try{i(a.next(t))}catch(t){e(t)}}function s(t){try{i(a.throw(t))}catch(t){e(t)}}function i(e){var n;e.done?t(e.value):(n=e.value,n instanceof o?n:new o((function(t){t(n)}))).then(r,s)}i((a=a.apply(n,c||[])).next())}));var n,c,o,a}}};