"use strict";exports.id=6564,exports.ids=[6564],exports.modules={"./src/commands/disable.ts"(e,s,r){r.r(s),r.d(s,{handleMcpDisable:()=>d});var a=r("../../../../../../../../../../private/tmp/buildkite-job-01a03ad4-00a2-4e4f-9a4f-0e1ec809f684/agent-cli-mac-macos-x64-01a03ad4-00a2-4e4f-9a4f-0e1ec809f684.LshlV881/pnpm-virtual-store/chalk@5.6.2/node_modules/chalk/source/index.js"),t=r("./src/console-io.ts"),o=r("./src/mcp/disabled-store.ts"),i=r("./src/mcp/project-paths.ts");function d(e,s){return r=this,d=void 0,c=function*(){var e;try{const e=process.cwd(),{disabledPath:r}=(0,i.C)(e),d=new o.P(r),n=yield d.isServerDisabled(s);n||(yield d.addDisabledServer(s)),n?(0,t.OT)(a.Ay.yellow(`MCP server '${s}' was already disabled`)):(0,t.OT)(a.Ay.green(`✓ Disabled MCP server: ${s}`)+"\n"+a.Ay.dim("The server will no longer be loaded or prompted for approval.")),process.exit(0)}catch(s){const r=null!==(e=null==s?void 0:s.message)&&void 0!==e?e:String(s);(0,t.p2)(a.Ay.red(`Failed to disable MCP server: ${r}`)),process.exit(1)}},new((n=void 0)||(n=Promise))((function(e,s){function a(e){try{o(c.next(e))}catch(e){s(e)}}function t(e){try{o(c.throw(e))}catch(e){s(e)}}function o(s){var r;s.done?e(s.value):(r=s.value,r instanceof n?r:new n((function(e){e(r)}))).then(a,t)}o((c=c.apply(r,d||[])).next())}));var r,d,n,c}}};