"use strict";exports.id=4199,exports.ids=[4199],exports.modules={"./src/commands/repo-status.tsx"(t,n,e){e.r(n),e.d(n,{handleRepoStatus:()=>s});var o=e("./src/console-io.ts");function s(t){return n=this,e=void 0,r=function*(){try{const n=yield t.getIndexingStatus();(0,o.OT)(JSON.stringify(n,null,2))}catch(t){(0,o.p2)(t instanceof Error?t.message:String(t))}process.exit(0)},new((s=void 0)||(s=Promise))((function(t,o){function i(t){try{a(r.next(t))}catch(t){o(t)}}function c(t){try{a(r.throw(t))}catch(t){o(t)}}function a(n){var e;n.done?t(n.value):(e=n.value,e instanceof s?e:new s((function(t){t(e)}))).then(i,c)}a((r=r.apply(n,e||[])).next())}));var n,e,s,r}}};