export declare const Errors: {
    ThreadTermination: () => Error;
    FilenameNotProvided: () => Error;
    TaskQueueAtLimit: () => Error;
    NoTaskQueueAvailable: () => Error;
    CloseTimeout: () => Error;
};
