"use strict";exports.id=4343,exports.ids=[4343],exports.modules={"../cursor-sdk-local-runtime/dist/run-store/sqlite/node-sqlite-driver.js"(e,t,s){s.r(t),s.d(t,{openNodeSqliteDriver:()=>i});var r=s("node:sqlite");class n{constructor(e){this.db=e,this.statements=new Map}exec(e){this.db.exec(e)}run(e,t=[]){const s=this.prepare(e).run(...t);return{changes:Number(s.changes)}}get(e,t=[]){return this.prepare(e).get(...t)}all(e,t=[]){return this.prepare(e).all(...t)}close(){this.statements.clear(),this.db.close()}prepare(e){let t=this.statements.get(e);return t||(t=this.db.prepare(e),this.statements.set(e,t)),t}}function i(e){return new n(new r.DatabaseSync(e))}}};