"use strict";exports.id=8914,exports.ids=[8914],exports.modules={"../cursor-sdk-local-runtime/dist/run-store/sqlite/bun-sqlite-driver.js"(e,t,s){s.r(t),s.d(t,{openBunSqliteDriver:()=>i});var r=s("bun:sqlite");class n{constructor(e){this.db=e,this.statements=new Map}exec(e){this.db.run(e)}run(e,t=[]){const{changes:s}=this.prepare(e).run(...t);return{changes:Number(s)}}get(e,t=[]){const s=this.prepare(e).get(...t);return null!=s?s:void 0}all(e,t=[]){return this.prepare(e).all(...t)}close(){this.statements.clear(),this.db.close()}prepare(e){let t=this.statements.get(e);return t||(t=this.db.prepare(e),this.statements.set(e,t)),t}}function i(e){return new n(new r.Database(e,{create:!0}))}}};