"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.commonState = void 0;
exports.commonState = {
    isWorkerThread: false,
    workerData: undefined
};
//# sourceMappingURL=types.js.map