"use strict";exports.id=1449,exports.ids=[1449],exports.modules={"./src/commands/repo-info.tsx"(n,t,o){o.r(t),o.d(t,{handleRepoInfo:()=>r});var e=o("./src/console-io.ts");function r(n){return t=this,o=void 0,i=function*(){try{const t=yield n.getRepositoryInfo();(0,e.OT)(JSON.stringify(t,null,2))}catch(n){(0,e.p2)(n instanceof Error?n.message:String(n))}process.exit(0)},new((r=void 0)||(r=Promise))((function(n,e){function s(n){try{a(i.next(n))}catch(n){e(n)}}function c(n){try{a(i.throw(n))}catch(n){e(n)}}function a(t){var o;t.done?n(t.value):(o=t.value,o instanceof r?o:new r((function(n){n(o)}))).then(s,c)}a((i=i.apply(t,o||[])).next())}));var t,o,r,i}}};