"use strict";exports.id=7679,exports.ids=[7679],exports.modules={"./src/commands/install-shell-integration.ts"(t,e,n){n.r(e),n.d(e,{handleInstallShellIntegration:()=>c});var l=n("node:fs"),s=n("node:os"),i=n("node:path"),r=n.n(i),o=n("./src/console-io.ts");const a='eval "$(~/.local/bin/agent shell-integration zsh)"';function c(){const t=r().join((0,s.homedir)(),".zshrc");try{let e="";try{e=(0,l.readFileSync)(t,"utf8")}catch(t){if("ENOENT"!==t.code)throw t;e=""}if(e.includes(a)||e.includes('eval "$(~/.local/bin/cursor-agent shell-integration zsh)"'))return void(0,o.p2)("Shell integration is already installed in ~/.zshrc");const n=`${a}\n${e}`;(0,l.writeFileSync)(t,n,"utf8"),(0,o.p2)("Shell integration installed successfully!"),(0,o.p2)("Please restart your terminal or run 'source ~/.zshrc' to activate.")}catch(t){(0,o.p2)(`Failed to install shell integration: ${String(t)}`),process.exit(1)}}}};