"use strict";exports.id=9086,exports.ids=[9086],exports.modules={"./src/commands/uninstall-shell-integration.ts"(t,n,e){e.r(n),e.d(n,{handleUninstallShellIntegration:()=>d});var i=e("node:fs"),l=e("node:os"),s=e("node:path"),o=e.n(s),r=e("./src/console-io.ts");const a='eval "$(~/.local/bin/agent shell-integration zsh)"',c='eval "$(~/.local/bin/cursor-agent shell-integration zsh)"';function d(){const t=o().join((0,l.homedir)(),".zshrc");try{let n="";try{n=(0,i.readFileSync)(t,"utf8")}catch(t){if("ENOENT"===t.code)return void(0,r.p2)("~/.zshrc file not found. Shell integration is not installed.");throw t}if(!n.includes(a)&&!n.includes(c))return void(0,r.p2)("Shell integration is not installed in ~/.zshrc");const e=n.split("\n"),l=e.filter((t=>t.trim()!==a&&t.trim()!==c));e[0]!==a&&e[0]!==c||""!==e[1]||l.shift();const s=l.join("\n");(0,i.writeFileSync)(t,s,"utf8"),(0,r.p2)("Shell integration uninstalled successfully!"),(0,r.p2)("Please restart your terminal for changes to take effect.")}catch(t){(0,r.p2)(`Failed to uninstall shell integration: ${String(t)}`),process.exit(1)}}}};