import mod from "./main.js";

export default mod;
export const ArrayTaskQueue = mod.ArrayTaskQueue;
export const FixedQueue = mod.FixedQueue;
export const Piscina = mod.Piscina;
export const isWorkerThread = mod.isWorkerThread;
export const move = mod.move;
export const queueOptionsSymbol = mod.queueOptionsSymbol;
export const transferableSymbol = mod.transferableSymbol;
export const valueSymbol = mod.valueSymbol;
export const version = mod.version;
export const workerData = mod.workerData;
