#ifndef NODE_TREE_SITTER_LANGUAGE_H_
#define NODE_TREE_SITTER_LANGUAGE_H_

#include "tree_sitter/api.h"

#include <napi.h>
#include <node_object_wrap.h>

namespace node_tree_sitter::language_methods {

void Init(Napi::Env env, Napi::Object);

const TSLanguage *UnwrapLanguage(Napi::Value);

} // namespace node_tree_sitter::language_methods

#endif // NODE_TREE_SITTER_LANGUAGE_H_
