import type { TaskQueue, Task } from '.';
export declare class ArrayTaskQueue implements TaskQueue {
    tasks: Task[];
    get size(): number;
    shift(): Task | null;
    push(task: Task): void;
    remove(task: Task): void;
}
